
// frontline.l                           cellular automata paint program
// by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
//
// ??-??-??: wrote many years ago...
// 09-11-00: first released version

lib_include "Lib/lib_home.h";
lib_include "Win/win.l";

int done,leftdown,addn;
win w;

#define NAT 3
#define N 64

char *picname = "Funstuff/Frontline/marballaI.tif";
pic p;

double civ[N][N][NAT];     // civilians (current)
double mil[N][N][NAT];     // military (current)
double civ1[N][N][NAT];    // civilians (new)
double mil1[N][N][NAT];    // military (new)

void wevent(win w,int eventnum);
void do_draw();
void do_step();
double fc(int i,int j,int n);
void addciv(int n);
void randciv();
void picciv();

lib_main {
    int x,y,sx,sy;
    int i,j,n;
    char name[256];

    sprintf(name,"%s/%s",LIB_LIBDIR,picname);
    p=pic_read(name);
    if (!p) exit(1);
    for (i=0;i<N;i++) {
        for (j=0;j<N;j++) {
            for (n=0;n<NAT;n++) {
                civ[i][j][n]=0.0;
                mil[i][j][n]=0.0;
            }
        }
    }
    civ[5][N-6][0]=1.0;
    civ[N-6][5][1]=1.0;
    civ[N-20][n-20][2]=1.0;

    win_screensize(&sx,&sy);
    x=y=768;
    leftdown=0; addn=0;
    w=win_create("frontline",sx/2-x/2,sy/2-y/2,x,y,wevent);
    win_repeat();
    for (done=0;!done;) sel_wait(300000);
};

void wevent(win w,int eventnum)
{
    switch (eventnum) {
    case WIN_QUIT:
        done=1;
        break;
    case WIN_DRAW:
        do_draw();
        break;
    case WIN_KEY:
        switch (w->key) {
        case 0x1b:
            done=1;
            break;
        case '0':
        case '1':
        case '2':
            addn=w->key-'0';
            break;
        case 'r':
            randciv();
            break;
        case 'p':
            picciv();
            break;
        default:
            break;
        }
        break;
    case WIN_REPEAT:
        do_step();
        do_draw();
        break;
    case WIN_LEFTDOWN:
        leftdown=1;
        break;
    case WIN_LEFTUP:
        leftdown=0;
        break;
    default:
        break;
    }
}

void do_draw()
{
    int i,j,n,x,y,d,r,g,b;
    double a;

    d=768/N;
    for (i=0;i<N;i++) {
        for (j=0;j<N;j++) {
            a=civ[i][j][0];
            r=(int)floor(255.99*a/2.0); if (r<0) r=0; if (r>255) r=255;
            a=civ[i][j][1];
            g=(int)floor(255.99*a/2.0); if (g<0) g=0; if (g>255) g=255;
            a=civ[i][j][2];
            b=(int)floor(255.99*a/2.0); if (b<0) b=0; if (b>255) b=255;
            win_color(r,g,b);
            x=i*d; y=j*d;
            win_rectf(x,y,d,d);
        }
    }
}

void do_step()
{
    int i,j,n;
    double a,q,b,c,u,s;

    if (leftdown) addciv(addn);
    for (i=0;i<N;i++) {
        for (j=0;j<N;j++) {
            for (n=0;n<NAT;n++) {
                a=(fc(i-1,j,n)+fc(i+1,j,n)+fc(i,j-1,n)+fc(i,j+1,n))/4.0;
                q=0.8;
                a=q*civ[i][j][n]+(1.0-q)*a;
                if (a>1e-5 && a<1.0) a*=1.05; if (a>2.0) a*=0.95;
                q=RND; q=1.0+0.001*q; a*=q;
                civ1[i][j][n]=a;
            }
        }
    }
    u=0.2;
    for (i=0;i<N;i++) {
        for (j=0;j<N;j++) {
            a=civ1[i][j][0];
            b=civ1[i][j][1];
            c=civ1[i][j][2];
            s=a+b+c;
            if (s>1e-5) {
                q=a/s; q=u*q+(1.0-u); civ1[i][j][0]*=q;
                q=b/s; q=u*q+(1.0-u); civ1[i][j][1]*=q;
                q=c/s; q=u*q+(1.0-u); civ1[i][j][2]*=q;
            }
        }
    }
    for (i=0;i<N;i++) {
        for (j=0;j<N;j++) {
            for (n=0;n<NAT;n++) {
                civ[i][j][n]=civ1[i][j][n];
            }
        }
    }
}

double fc(int i,int j,int n)
{
    if (i<0) i=0; if (i>=N) i=N-1;
    if (j<0) j=0; if (j>=N) j=N-1;
    return civ[i][j][n];
}

void addciv(int n)
{
    int x,y,i,j,d;

    win_mouse(&x,&y);
    d=768/N;
    i=x/d; if (i<0) i=0; if (i>=N) i=N-1;
    j=y/d; if (j<0) j=0; if (j>=N) j=N-1;
    civ[i][j][n]+=5.0;
}

void randciv()
{
    int i,j,n;
    double q;

    for (i=0;i<N;i++) {
        for (j=0;j<N;j++) {
            for (n=0;n<NAT;n++) {
                q=RND; q=0.2*(q+1.0); if (q<0.0) q=0.0;
                civ[i][j][n]=q;
            }
        }
    }
}

void picciv()
{
    int i,j,n;
    double q;

    for (i=0;i<N;i++) {
        for (j=0;j<N;j++) {
            for (n=0;n<NAT;n++) {
                q=(double)((p->raster[j*p->xsize+i]>>(n*8))&0xff)/127.9;
//...more
                if (n==2) q*=0.75; if (n==1) q*=1.3;
                civ[i][j][n]=q;
            }
        }
    }
}


