
#define PIC__SOURCE 1

lib_public {

    // pic.l                                  multiple-format picture support
    // by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
    //
    // 08-26-94: wrote
    // 12-07-94: picform's and events
    // 09-21-95: added ras format(s)
    // 11-27-96: allowed _iconify() to expand; _sample0/1()
    // 07-16-97: (a) moved applications picshow,picdiff to Win to keep Pic
    //           compiles from depending on Win
    //           (b) added picevent support, only _DESTROY callback for now
    // 10-15-99: added lib_genexec/lib_genclean calls to handle Tiff/Jpeg
    // 10-17-99: MAJOR CHANGE--went from ABGR raster to RGBA raster
    //           NOTE that lot's of tools will need fixing
    // 03-17-00: added SGI .rgb format
    // 09-08-00: added autobuild of Mpeg utility
    // 09-11-00: split off configjpeg script activity to get jconfig.h early
    // 11-03-00: added picblend.l utility
    // 12-13-00: added uncolormap.l utility

    // perform make/make clean on attached libraries/utilities
    lib_configexec "./configjpeg";
    lib_genexec "./genjpeg";
    lib_genexec "./gentiff";
    lib_genexec "./genmpeg";
    lib_cleanexec "./cleanjpeg";
    lib_cleanexec "./cleantiff";
    lib_cleanexec "./cleanmpeg";

    lib_include "Sys/sys.l";

    #define PICFORM_BUFFERCHECK      0
    #define PICFORM_NAMECHECK        1
    #define PICFORM_DECODE           2
    #define PICFORM_ENCODE           3
    #define PICFORM_ENCODEARGS       4

    typedef char *(*picformevent)(int eventnum,char *data);
    typedef unsigned char picquan;

    lib_struct {
        picformevent event;
        picform pf0,pf1;
    } picform;

    #ifndef PIC__SOURCE
        extern picform picform0,picform1;
    #endif

    void picform_create(picformevent event);
    char *picform_suffixcheck(char *name,char *suffix);

    #define PIC_EVENT_DESTROY   0

    typedef void (*picevent)(pic p,int eventnum);

    lib_struct {
        int xsize,ysize;                // pixels per axis
        picquan *pbuf;                  // pixel storage (GL_RGBA/ubyte)
        int *raster;                    // alternate pointer to pbuf
        picevent event;                 // application-specific event handler
        char *data;                     // application-specific data
    } pic;

    pic pic_create(int xsize,int ysize,int *raster);
    void pic_destroy(pic p);
    pic pic_decode(char *buf);
    char *pic_encode(pic p,char *name);
    pic pic_read(char *name);
    int pic_write(pic p,char *name);
    pic pic_iconify(pic p,int dxsize,int dysize,
        double xstretch,int sx0,int sy0,int sxsize,int sysize);
    void pic_resize(pic ps,pic pd);
    int pic_sample0(pic p,int x,int y);
    int pic_sample1(pic p,double x,double y);

    lib_archive {
        "Gif/gif.l",
        "Jpg/jpg.l",
        "Pnm/pnm.l",
        "Ras/ras.l",
        "Rgb/rgb.l",
        "Tif/tif.l"
    };

};


lib_application "Pic2mpg/pic2mpg.l";
lib_application "Pic2ps/pic2ps.l";
lib_application "Pic2vidpic/pic2vidpic.l";
lib_application "Picblend/picblend.l";
lib_application "Picconvert/picconvert.l";
lib_application "Piciconify/piciconify.l";
lib_application "Picpaste/picpaste.l";
lib_application "Rgbcombine/rgbcombine.l";
lib_application "Rndpic/rndpic.l";
lib_application "Uncolormap/uncolormap.l";


picform picform0=(picform)0,picform1=(picform)0;
char *pic_buf=(char *)0;
int pic_bufsize= -1;


void picform_create(picformevent event)
{
    picform pf;

    pf=(picform)mem_get(sizeof(picform_struct),"picform_create");
    pf->event=event;
    pf->pf0=picform1; pf->pf1=(picform)0;
    if (pf->pf0) pf->pf0->pf1=pf; else picform0=pf;
    if (pf->pf1) pf->pf1->pf0=pf; else picform1=pf;
}

char *picform_suffixcheck(char *name,char *suffix)
{
    int i,i0;

    for (i=0,i0=0;name[i];i++) if (name[i]=='.') i0=i+1;
    name+=i0;
    while (*suffix && *name && (*suffix=='.' || ((*suffix^*name)&~0x20)==0))
        { suffix++; name++; }
    if (*suffix) return (char *)1;
    else return (char *)0;
}


pic pic_create(int xsize,int ysize,int *raster)
{
    pic p;
    int *r,*s,*e;

    p=(pic)mem_get(sizeof(pic_struct),"pic_create");
    p->xsize=xsize; p->ysize=ysize;
    mem_noclear();
    p->pbuf=(picquan *)mem_get(xsize*ysize*4,"pic_create");
    p->raster=(int *)(p->pbuf);
    if (raster) {
        for (r=p->raster,e=r+xsize*ysize,s=raster;r<e;)
            *r++ = *s++;
    }else{
        // for (r=p->raster,e=r+xsize*ysize;r<e;) *r++ =0x00000000;
    }
    p->event=(picevent)0; p->data=(char *)0;
    return p;
}

void pic_destroy(pic p)
{
    if (p->event) (*p->event)(p,PIC_EVENT_DESTROY);
    mem_put((char *)p->pbuf,"pic_destroy");
    mem_put((char *)p,"pic_destroy");
}

// NOTE: currently _decode and the DECODE event handlers assume
// that buf was obtained from a mem_get() call so that mem_size(buf)
// doesn't fail.

pic pic_decode(char *buf)
{
    picform pf;
    pic p;
    char *s;

    if (mem_size(buf)<20) return (pic)0;
    for (pf=picform0;pf;pf=pf->pf1) {
        s=(*pf->event)(PICFORM_BUFFERCHECK,buf);
        if (s==(char *)0) {
            p=(pic)(*pf->event)(PICFORM_DECODE,buf);
            if (p) return p;
        }
    }
    return (pic)0;
}

// NOTE: _encode() currently uses mem_get() to allocate the return buf.
// This means that mem_size(buf) can be used to determine the number of
// encoded bytes.

char *pic_encode(pic p,char *name)
{
    char *buf;
    picform pf;

    for (pf=picform0;pf;pf=pf->pf1) {
        if ((*pf->event)(PICFORM_NAMECHECK,name)==(char *)0 &&
            (buf=(*pf->event)(PICFORM_ENCODE,(char *)p))) return buf;
    }
    return (char *)0;
}

pic pic_read(char *name)
{
    char *buf;
    pic p;

    buf=fio_read(name);
    if (!buf) return (pic)0;
    p=pic_decode(buf);
    mem_put(buf,"pic_read");
    return p;
}

int pic_write(pic p,char *name)
{
    int r;
    char *buf;

    if (!(buf=pic_encode(p,name))) return -1;
    r=fio_write(name,buf,mem_size(buf));
    mem_put(buf,"pic_write");
    return r;
}

pic pic_iconify(pic p,int dxsize,int dysize,
    double xstretch,int sx0,int sy0,int sxsize,int sysize)
{
    int i,j,k,c,ci,cj,xsize,ysize,*raster;
    double xscale,yscale,scale;
    int *rsum,*gsum,*bsum,*tsum;
    int *craster,*imap,*jmap,*s;
    picquan *q;
    pic pd;

    xsize=p->xsize; ysize=p->ysize; raster=p->raster;

    // allocate temporary and result memory
    rsum=(int *)mem_get(dxsize*dysize*sizeof(int),"iconify");
    gsum=(int *)mem_get(dxsize*dysize*sizeof(int),"iconify");
    bsum=(int *)mem_get(dxsize*dysize*sizeof(int),"iconify");
    tsum=(int *)mem_get(dxsize*dysize*sizeof(int),"iconify");
    for (cj=0;cj<dysize;cj++) {
        for (ci=0;ci<dxsize;ci++) {
            k=cj*dxsize+ci;
            rsum[k]=0; gsum[k]=0; bsum[k]=0; tsum[k]=0;
        }
    }
    imap=(int *)mem_get(sxsize*sizeof(int),"iconify");
    jmap=(int *)mem_get(sysize*sizeof(int),"iconify");
    pd=pic_create(dxsize,dysize,(int *)0);
    craster=pd->raster;

    // get axis scale factors
    xscale=(double)dxsize/(xstretch*(double)sxsize);
    yscale=(double)dysize/(double)sysize;
    scale=(xscale<yscale)?xscale:yscale;

    // deal with expand/contract cases
    if (scale<=0.707) {
        // contraction: put source pixels into destination buckets and
        // take bucket averages as result

        // make transform tables
        if (scale*xstretch>1.0) scale=1.0/xstretch;
        for (i=0;i<sxsize;i++)
            imap[i]=
                (int)floor(((double)i+0.5-(double)sxsize/2.0)*scale*xstretch+
                (double)dxsize/2.0 );
        for (j=0;j<sysize;j++)
            jmap[j]=(int)floor(((double)j+0.5-(double)sysize/2.0)*scale+
                (double)dysize/2.0 );
    
        // total up contributions
        for (j=sy0;j<sy0+sysize;j++) {
            s=raster+j*xsize;
            for (i=sx0;i<sx0+sxsize;i++) {
                ci=imap[i-sx0]; cj=jmap[j-sy0];
                k=cj*dxsize+ci;
                q=(picquan *)(s+i);
                rsum[k]+= *q++;
                gsum[k]+= *q++;
                bsum[k]+= *q++;
                tsum[k]++;
            }
        }

        // put result in icon
        for (cj=0;cj<dysize;cj++) {
            for (ci=0;ci<dxsize;ci++) {
                int r,g,b;
    
                k=cj*dxsize+ci;
                c=tsum[k];
                if (c) {
                    r=rsum[k]/c; g=gsum[k]/c; b=bsum[k]/c;
                }else{
                    r=  0; g=  0; b=  0;
                }
                q=(picquan *)(craster+k);
                *q++ =r; *q++ =g; *q++ =b; *q++ =0;
            }
        }
    }else{
        // expansion: sample bilinearly interpolated source
        // per destination pixel
        for (cj=0;cj<dysize;cj++) {
            for (ci=0;ci<dxsize;ci++) {
                craster[cj*dxsize+ci]=pic_sample1(p,
                    (double)sx0+(double)ci/xscale,
                    (double)sy0+(double)cj/yscale);
            }
        }
    }
    

    // free up temporary storage
    mem_put((char *)imap,"iconify"); mem_put((char *)jmap,"iconify");
    mem_put((char *)rsum,"iconify"); mem_put((char *)gsum,"iconify");
    mem_put((char *)bsum,"iconify"); mem_put((char *)tsum,"iconify");

    return pd;
}

void pic_resize(pic ps,pic pd)
{
    int k,r,g,b;
    int sx,sy,sxsize,sysize,*sras;
    int dx,dy,dxsize,dysize,*dras;
    double scale,sxc,syc,dxc,dyc;

    sxsize=ps->xsize; sysize=ps->ysize; sras=ps->raster;
    dxsize=pd->xsize; dysize=pd->ysize; dras=pd->raster;
    scale=((double)sxsize/(double)dxsize+(double)sysize/(double)dysize)/2.0;
    sxc=(double)sxsize/2.0; syc=(double)sysize/2.0;
    dxc=(double)dxsize/2.0; dyc=(double)dysize/2.0;
    for (dy=0;dy<dysize;dy++) {
        for (dx=0;dx<dxsize;dx++) {
            sx=(int)floor(((double)dx-dxc)*scale+sxc+0.5);
            sy=(int)floor(((double)dy-dyc)*scale+syc+0.5);
            if (sx<0 || sx>=sxsize || sy<0 || sy>=sysize) k=0;
            else k=sras[sy*sxsize+sx];
            *dras++ =k;
        }
    }
}

int pic_sample0(pic p,int x,int y)
{
    int xsize,ysize;

    xsize=p->xsize; ysize=p->ysize;
    if (x<0 || x>=xsize || y<0 || y>=ysize) return 0;
    return p->raster[y*xsize+x];
}

#define SAMP1SUB(I,J,R,G,B) {                                                 \
    *kp=pic_sample0(p,I,J);                                                   \
    R=(double)(kbuf[0]); G=(double)(kbuf[1]); B=(double)(kbuf[2]);            \
}

int pic_sample1(pic p,double x,double y)
{
    int i,j,k,xsize,ysize,ri,gi,bi,q,*kp;
    double u,v,u1,v1;
    double r00,g00,b00,r01,g01,b01,r10,g10,b10,r11,g11,b11;
    double r0,g0,b0,r1,g1,b1;
    double r,g,b;
    picquan kbuf[4];

    i=(int)floor(x); u=x-(double)i; u1=1.0-u;
    j=(int)floor(y); v=y-(double)j; v1=1.0-v;
    kp=(int *)(&kbuf[0]);
    SAMP1SUB(i  ,j  ,r00,g00,b00) SAMP1SUB(i+1,j  ,r10,g10,b10)
    SAMP1SUB(i  ,j+1,r01,g01,b01) SAMP1SUB(i+1,j+1,r11,g11,b11)
    r0=u1*r00+u*r10; g0=u1*g00+u*g10; b0=u1*b00+u*b10;
    r1=u1*r01+u*r11; g1=u1*g01+u*g11; b1=u1*b01+u*b11;
    r=v1*r0+v*r1; g=v1*g0+v*g1; b=v1*b0+v*b1;
    q=255.99999/255.0;
    kbuf[0]=(int)floor(q*r); kbuf[1]=(int)floor(q*g); kbuf[2]=(int)floor(q*b);
    kbuf[3]=0;
    return *kp;
}

