
// spintest.l                                    test spin some geometry
// by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
//
// 06-07-99: wrote
// 10-19-99: split spintest.l off when spin.l made into library
// 02-22-00: DO_PALETTE compile-time option; changed so ./texture.ppm becomes
//           $LIB_LIBDIR/Data/LibGenTest/spintest_texture.ppm
// 07-06-00: texture .ppm to .jpg
// 01-30-01: use of texture id to be compatible with win_ogl.l etc...


lib_include "Lib/lib_home.h";
lib_include "../spin.l";

// for paletted-texture test, change sPINTEST_DO_PALETTE to SPINTEST_DO_PALETTE
#define sPINTEST_DO_PALETTE

#define N 100
//#define N 224
//#define N 400
#define N2 (2*N)

spin sp;
int done,dolist;
pic p_tex;
float uvtab[N][N][2];
float xyztab[N][N][3];
double t0,t;
int frames;
GLuint tex_id;

#ifdef SPINTEST_DO_PALETTE
unsigned char *p_tex1;
unsigned char rgbatab[1024];
#endif

void send_strips();

void spevent(spin sp,int e);
void do_draw();
void send_strips();

lib_main {
    int i,j,x,y,size;

    size=976; dolist=1;
    win_screensize(&x,&y);
    sp=spin_create("spintest",x/2-size/2,y/2-size/2,size,size,spevent);

    frames=0;

#ifdef SPINTEST_DO_PALETTE
for (i=0;i<256;i++) {
    rgbatab[4*i+0]=0;
    rgbatab[4*i+1]=255-i;
    rgbatab[4*i+2]=0;
    rgbatab[4*i+3]=0;
}
#endif


    {
        int *ras,m;
        unsigned char kc[4];
#ifdef SPINTEST_DO_PALETTE
unsigned char *pt1,*ras1;
#endif

        kc[0]=0; kc[1]=0; kc[2]=0; kc[3]=0xff; ras=(int *)(&kc[0]); m= *ras;
        {
            char texfile[1024];

            sprintf(texfile,"%s/data/spintest_texture.jpg",
                LIB_LIBDIR);
            p_tex=pic_read(texfile);
            if (!p_tex) {
                fprintf(stderr,"spintest: can't read %s\n",texfile); exit(1);
            }
        }
        x=p_tex->xsize; y=p_tex->ysize;
        ras=p_tex->raster;
        for (j=0;j<y;j++)
            { for (i=0;i<x;i++) (*ras++)|=m; }
#ifdef SPINTEST_DO_PALETTE
{
    double r,g;
    int m0;

    p_tex1=(unsigned char *)mem_get(x*y,"main");
    pt1=p_tex1; ras1=(unsigned char *)p_tex->raster;
    m0=0;
    for (j=0;j<y;j++) {
        for (i=0;i<x;i++) {
            r=(double)(*ras1++);
            g=(double)(*ras1++);
            ras1+=2;
            if (r+g>4.0) m=m0=(int)floor(255.999999*r/(r+g)+0.0000005);
            else m=m0;
            *pt1++ =m;
        }
    }
}
#endif


    glGenTextures(1,&tex_id);
    glBindTexture(GL_TEXTURE_2D,tex_id);
#ifdef SPINTEST_DO_PALETTE
        glTexImage2D(GL_TEXTURE_2D,0,GL_LUMINANCE,
            p_tex->xsize,p_tex->ysize,0,GL_LUMINANCE,
            GL_UNSIGNED_BYTE,(GLvoid *)p_tex1);
#else
        glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,
            p_tex->xsize,p_tex->ysize,0,GL_RGBA,
            GL_UNSIGNED_BYTE,(GLvoid *)p_tex->raster);
#endif
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
    }

#ifdef SPINTEST_DO_PALETTE
#ifdef GL_SHARED_TEXTURE_PALETTE_EXT
glColorTableSGI(GL_SHARED_TEXTURE_PALETTE_EXT, GL_RGBA,
    256, GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid *)rgbatab);
glEnable(GL_SHARED_TEXTURE_PALETTE_EXT);
#else
#ifdef GL_TEXTURE_COLOR_TABLE_SGI
glColorTableSGI(GL_TEXTURE_COLOR_TABLE_SGI, GL_RGBA,
    256, GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid *)rgbatab);
glEnable(GL_TEXTURE_COLOR_TABLE_SGI);
#endif
#endif
#endif


    {
        double x,y,z,a,r,d,q,tc0,tc1;

        r=0.9; d=0.8*r;
        tc0=0.5/(double)p_tex->xsize;
        tc1=(double)(p_tex->xsize-1)/(double)p_tex->xsize;
        for (i=0;i<N;i++) {
            for (j=0;j<N;j++) {
                a=2.0*PI*(double)j/(double)(N-1);
                z= -d+2.0*d*(double)i/(double)(N-1);
                q=sqrt(r*r-z*z);
                x=q*cos(a); y=q*sin(a);
                xyztab[i][j][0]=x; xyztab[i][j][1]=y; xyztab[i][j][2]=z;
                uvtab[i][j][0]=tc0+tc1*(double)i/(double)(N-1);
                uvtab[i][j][1]=tc0+tc1*(double)j/(double)(N-1);
            }
        }
    }

    glNewList(1,GL_COMPILE);
    send_strips();
    glEndList();

    for (done=0;!done;) sel_wait(300000);
};

void spevent(spin sp,int e)
{
    int i,j;
    win w;

    w=sp->w;
    switch (e) {
    case WIN_KEY:
        switch (w->key) {
        case 0x1b: done=1; break;
        case 'l': dolist=1; break;
        case 'L': dolist=0; break;
        case '?':
            for (i=0;i<4;i++) {
                for (j=0;j<4;j++) printf(" %19.13f",sp->mview->m[i][j]);
                printf("\n");
            }
            break;
        default:
            break;
        }
        break;
    case WIN_DRAW:
        do_draw();
        break;
    case WIN_QUIT:
        done=1;
        break;
    default:
        break;
    }
}

void do_draw()
{
    if (dolist) glCallList(1);
    else send_strips();

    frames++;
}

void send_strips()
{
    int i,j;

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex_id);

#define VERTOUT(I,J) { glTexCoord2fv(uvtab[I][J]); glVertex3fv(xyztab[I][J]); }

    for (i=0;i+1<N;i++) {
        glBegin(GL_TRIANGLE_STRIP);
        for (j=0;j<N;j++) { VERTOUT(i,j) VERTOUT(i+1,j) }
        glEnd();
    }

    glDisable(GL_TEXTURE_2D);
}


