
lib_public {

    // spin.l                                    spin geometry in a window
    // by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
    //
    // 06-07-99: wrote
    // 10-15-99: made into a library module
    // 12-17-00: added _setmodelvmat()
    // 12-23-00: actually use name in spin_create()
    // 01-30-01: new 'a'/'z' key interpretation: acceleration
    // 01-31-01: moved 'w' screen dump keystroke to be common for all apps;
    //           added 'f'/'F' for frames-per-second meter


    lib_include "Geowin/geowin.l"; // for win, view and 4x4 view-matrix stuff


    // NOTE: application callback gets events very similar to win events,
    // except some are caught by spin and not passed on and the first
    // callback arg is a spin object.  Take eventnum usage from win.l.
    // Spin windows are automatically double-buffered and
    // win_repeat/norepeated, so don't try to use those commands.  It is
    // safest to only read window state and perform drawing commands,
    // and avoiding making changes to the win directly (use or create
    // spin_...() wrappers in general for actions modifying win state).

    typedef void (*spinevent)(spin sp,int eventnum);

    lib_struct {
        spinevent event;             // application callback for events
        char *data;                  // application-supplied attachment
        win w;                       // owned by spin
        view v;                      // owned by spin
        int dragmode;                // drag state
        double v_x0,v_y0,v_x1,v_y1;  // more drag state
        double speed;                // eyepoint speed
        double px,py,pz;             // eyepoint
        double px0,py0,pz0;          // start-drag eyepoint
        int mx0,my0;                 // start-drag mouse pos
        vmat mmodel,mmov,mrot,mper;  // view matrix components
        vmat mmodmov,mmrot,mview;    // view matrix aggregates
    } spin;

    spin spin_create(char *name,int x,int y,int xsiz,int ysiz,spinevent event);
    void spin_destroy(spin sp);
    void spin_setmodelvmat(spin sp,vmat mmodel);

    lib_archive { };

};


lib_application "Spintest/spintest.l";


void spin_wevent(win w,int e);
void spin_vevent(view v);
void spin_do_draw(spin sp);


spin spin_create(char *name,int x,int y,int xsiz,int ysiz,spinevent event)
{
    spin sp;

    sp=(spin)mem_get(sizeof(spin_struct),"spin_create");
    sp->event=event;
    sp->data=(char *)0;
    win_doublebuffer();
    sp->w=win_create(name,x,y,xsiz,ysiz,spin_wevent);
    sp->w->data=(char *)sp;
    win_repeat();
    sp->v=view_create((double)sp->w->xsize/2.0,(double)sp->w->ysize/2.0,
        (double)(sp->w->xsize+sp->w->ysize)/4.0,spin_vevent);
    sp->dragmode=0;
    sp->v_x0=0.0; sp->v_y0=0.0; sp->v_x1=0.0; sp->v_y1=0.0;
    sp->speed=0.0;
    sp->px=0.0; sp->py=1.5; sp->pz=0.0;
    sp->mmodmov=vmat_create(); sp->mmov=vmat_create(); sp->mrot=vmat_create();
    sp->mper=vmat_create(); sp->mmrot=vmat_create(); sp->mview=vmat_create();
    sp->mmodel=vmat_create(); vmat_ident(sp->mmodel);
    glClearColor(0.0,0.0,0.0,1.0);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    #ifdef GL_MULTISAMPLE_SGIS
        glDisable(GL_MULTISAMPLE_SGIS);
    #endif
    glDisable(GL_LIGHTING);
    glShadeModel(GL_FLAT);
    //glDisable(GL_TEXTURE_2D);
    glDisable(GL_CULL_FACE);
    glDrawBuffer(GL_BACK);
    glViewport(0,0,sp->w->xsize,sp->w->ysize);
    return sp;
}

void spin_destroy(spin sp)
{
    vmat_destroy(sp->mmodmov); vmat_destroy(sp->mmov); vmat_destroy(sp->mrot);
    vmat_destroy(sp->mper); vmat_destroy(sp->mmrot); vmat_destroy(sp->mview);
    vmat_destroy(sp->mmodel);
    view_destroy(sp->v); win_destroy(sp->w);
    mem_put((char *)sp,"spin_destroy");
}

void spin_setmodelvmat(spin sp,vmat mmodel)
{
    vmat_copy(mmodel,sp->mmodel);
}


// ========== internal routines ==========

void spin_wevent(win w,int e)
{
    int x,y;
    spin sp;
    view v;
    double d;

    sp=(spin)w->data; v=sp->v;
    win_mouse(&x,&y);
    switch (e) {
    case WIN_DRAW:
        x=(w->xsize+w->ysize)/2;
        view_resize((double)w->xsize/2.0,(double)w->ysize/2.0,(double)x/2.0);
        break;
    case WIN_LEFTDOWN: // spin object in front of viewer (actually, spin
                       // the whole world around a point 1 unit in front)
        if (sp->dragmode==0) {
            sp->dragmode=1;
            // ...more
        }
        break;
    case WIN_LEFTUP:
        if (sp->dragmode==1) { sp->dragmode=0; }
        break;
    case WIN_MIDDLEDOWN: // rotate view direction for flight
        if (sp->dragmode==0) {
            sp->dragmode=2;
            view_startdragrot((double)w->xsize-(double)x,
                (double)w->ysize-(double)y);
        }
        break;
    case WIN_MIDDLEUP:
        if (sp->dragmode==2) { sp->dragmode=0; }
        break;
    case WIN_RIGHTDOWN: // translate in plane of view
        if (sp->dragmode==0) {
            sp->dragmode=3;
            sp->px0=sp->px; sp->py0=sp->py; sp->pz0=sp->pz;
            sp->mx0=x; sp->my0=y;
        }
        break;
    case WIN_RIGHTUP:
        if (sp->dragmode==3) { sp->dragmode=0; }
        break;
    case WIN_KEY:
        d=sp->speed; if (d<0.0) d= -d; if (d<0.001) d=0.001;
        d*=sqrt(sqrt(sqrt(2.0)))-1.0;
        switch (w->key) {
        case 'A': sp->speed+=4.0*d; break;
        case 'a': sp->speed+=d; break;
        case 'z': sp->speed-=d; break;
        case 'Z': sp->speed-=4.0*d; break;
        case ' ': sp->speed=0.0; break;
        case 'w':
            {
                int xs,ys,mx,my;
                pic p;

                win_finish(); win_finishX();
                win_mouse(&mx,&my);
                xs=sp->w->xsize; ys=sp->w->ysize;
                p=pic_create(xs,ys,(int *)0);
                glReadPixels(0,0,xs,ys,GL_RGBA,GL_UNSIGNED_BYTE,
                    (GLvoid *)p->raster);
                pic_write(p,"spinout.tif");
                pic_destroy(p);
            }
            break;
        default:
            (*sp->event)(sp,e);
            break;
        }
        break;
    case WIN_REPEAT:
        d=sp->speed;
        sp->px-=d*sp->v->vx; sp->py-=d*sp->v->vy; sp->pz-=d*sp->v->vz;
        switch (sp->dragmode) {
        case 1:
            break;
        case 2:
            view_dragrot((double)w->xsize-(double)x,
                (double)w->ysize-(double)y);
            break;
        case 3:
            {
                double a,b;

                a=(double)(x-sp->mx0)/sp->v->scale;
                b=(double)(y-sp->my0)/sp->v->scale;
                sp->px=sp->px0+sp->v->ux*a+sp->v->wx*b;
                sp->py=sp->py0+sp->v->uy*a+sp->v->wy*b;
                sp->pz=sp->pz0+sp->v->uz*a+sp->v->wz*b;
            }
            break;
        default:
            break;
        }
/********************************************
 **  stuff for spin animation after leftup **
 ********************************************
        else{
            if (sp->leftup==1) {
                sp->v_x0=v->dragx; sp->v_y0=v->dragy;
                sp->v_x1=(double)x; sp->v_y1=(double)y;
                {
                    double dx,dy,r;

                    dx=sp->v_x1-sp->v_x0; dy=sp->v_y1-sp->v_y0;
                    r=sqrt(dx*dx+dy*dy);
                    sp->leftup=(r>1.9?2:0);
                }
            }
            if (sp->leftup==2) {
                v->dragx=sp->v_x0; v->dragy=sp->v_y0;
                view_dragrot(sp->v_x1,sp->v_y1);
            }
        }
********************************************/
        spin_do_draw(sp);
        break;
    default:
        (*sp->event)(sp,e);
        break;
    }
}

void spin_vevent(view v)
{
}

extern Display *win_display;

void spin_do_draw(spin sp)
{
    int i,j,k;
    double vm[16],qx,qy,q;
    win w;
    view v;

    w=sp->w; v=sp->v;

    q=(double)(w->xsize>=w->ysize?w->xsize:w->ysize);
    qx=q/(double)w->xsize;
    qy=q/(double)w->ysize;

    // fly-around mode
    vmat_move(sp->px,sp->py,sp->pz,sp->mmov);
    vmat_view(v,sp->mrot);
    //vmat_per(4.0,sp->mper);
    vmat_per(2.0,sp->mper);
    vmat_mul(sp->mmov,sp->mmodel,sp->mmodmov);
    vmat_mul(sp->mrot,sp->mmodmov,sp->mmrot);
    vmat_mul(sp->mper,sp->mmrot,sp->mview);

    // spin an object at the origin from (0,0,4) looking towards -z
    //vmat_view(v,sp->mrot);
    //vmat_move(0.0,0.0,4.0,sp->mmov);
    //vmat_per(4.0,sp->mper);
    //vmat_mul(sp->mmov,sp->mrot,sp->mmrot);
    //vmat_mul(sp->mper,sp->mmrot,sp->mview);

    glMatrixMode(GL_PROJECTION);
    for (k=0;k<16;k++) {
        i=k%4; j=k/4;
        if (i==0) vm[k]=qx*sp->mview->m[i][j];
        else{
            if (i==1) vm[k]=qy*sp->mview->m[i][j];
            else vm[k]=sp->mview->m[i][j];
        }
    }
    glLoadMatrixd(vm);

    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

    (*sp->event)(sp,WIN_DRAW);

    {
        static double t0=0.0;
        static double t1=0.0;
        static int ti= -1;
        static char ts[256];
        static double dt,dttab[10],dtsum;
        extern int win_font_tex_id[2];
        int x,y;

        t1=timer_systemdsec();
        if (t0!=0.0) {
            dt=t1-t0;
            if (ti<0) {
                for (ti=0,dtsum=0.0;ti<10;ti++) { dttab[ti]=dt; dtsum+=dt; }
                ti=0;
            }else{
                dtsum+=dt-dttab[ti]; dttab[ti]=dt; ti++; if (ti>=10) ti=0;
            }
            glLoadMatrixd(sp->w->impl->vm);
            glDisable(GL_DEPTH_TEST);
            sprintf(ts,"fps: %g",1.0/(dtsum/10.0));
            win_color(15,55,15);
            win_strsize(ts,&x,&y);
            win_rectf(10,10,x,y);
            win_color(255,0,255);
            win_str(10,10,ts);
            glEnable(GL_DEPTH_TEST);
        }
        t0=t1;
    }
    glXSwapBuffers(win_display,w->impl->xwin);
    glDrawBuffer(GL_BACK);
}


