
// surfcollapse.l                     utility front-end to ecol.l library
// by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
//
// 10-04-00: wrote

lib_include "Sys/sys.l";
lib_include "Surf/surf.l";

lib_main {
    int nv,ne,nf,op,i;
    double val;
    surf s;
    char *infile,*outfile;

    // get command-line arguments
    if (argc!=5) {
        printf("usage: surfop infile.mesb op val outfile.mesb\n");
        printf("    ops:\n");
        printf("      0 -- triangulate, val ignored\n");
        printf("      1 -- collapse, val=target vertex count\n");
        printf("      2 -- remove edges, val=flatness cutoff e.g. .5\n");
        printf("      3 -- subdivide, val: 1=linear, 2=quad, 3=C-C\n");
        exit(0);
    }

    infile=argv[1];
    if (sscanf(argv[2],"%d",&op)!=1) { fprintf(stderr,"bad arg\n"); exit(1); }
    if (sscanf(argv[3],"%lf",&val)!=1)
        { fprintf(stderr,"bad arg\n"); exit(1); }
    outfile=argv[4];

    s=surf_mesbread(infile);
    if (!s) { fprintf(stderr,"bad infile\n"); exit(1); }

    switch (op) {
    case 0:
        ecol_do_triangulate(s);
        break;
    case 1:
        ecol_do_collapse(s,(int)floor(val+0.5));
        break;
    case 2:
        ecol_do_remove(s,val);
        break;
    case 3:
        switch ((int)floor(val+0.5)) {
        case 1:
            surf_split1(s);
            break;
        case 2:
            surf_split2(s);
            break;
        case 3:
            surf_split3(s);
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }

    // write the result in the requested format
    printf("write the result in the requested format\n"); fflush(stdout);
    surf_size(s,&nv,&ne,&nf);
    printf("\nsurf size: nv=%d ne=%d nf=%d [mem=%d]\n",nv,ne,nf,mem_use());
    for (i=0;outfile[i];i++) ;
    i--;
    switch (outfile[i]) {
    case 'j':
        surf_macwrite(s,outfile);
        break;
    case 's':
        surf_write(s,outfile);
        break;
    case 'b':
        surf_mesbwrite(s,outfile);
        break;
    default:
        fprintf(stderr,"*** unknown format--not saving output\n");
        exit(1);
        break;
    }

};


